import nltk
import re
import string
from collections import Counter
import numpy as np
import pandas as pd
from bs4 import BeautifulSoup
from gensim.models import Phrases
import imblearn
from imblearn.combine import SMOTETomek
from imblearn.over_sampling import SMOTE, BorderlineSMOTE, SVMSMOTE
from imblearn.under_sampling import TomekLinks, NeighbourhoodCleaningRule
from nltk.corpus import wordnet
from nltk.stem import WordNetLemmatizer
from nltk.tokenize import TweetTokenizer
from sklearn.dummy import DummyClassifier
from sklearn.ensemble import RandomForestClassifier, AdaBoostClassifier
from sklearn.feature_extraction.stop_words import ENGLISH_STOP_WORDS
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.linear_model import SGDClassifier
from sklearn.metrics import confusion_matrix, precision_score, recall_score, f1_score
from sklearn.model_selection import GridSearchCV, cross_validate, cross_val_predict, KFold
from sklearn.naive_bayes import MultinomialNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.preprocessing import LabelEncoder
from sklearn.svm import SVC
from sklearn.tree import DecisionTreeClassifier

'''Display'''
from IPython.core.display import display, HTML

display(HTML("<style>.container { width:95% !important; }</style>"))
import warnings

warnings.filterwarnings('ignore')
pd.options.display.float_format = '{:,.2f}'.format

####

# Contraction map
c_dict = {
    "ain't": "am not",
    "aren't": "are not",
    "can't": "cannot",
    "can't've": "cannot have",
    "'cause": "because",
    "could've": "could have",
    "couldn't": "could not",
    "couldn't've": "could not have",
    "didn't": "did not",
    "doesn't": "does not",
    "don't": "do not",
    "hadn't": "had not",
    "hadn't've": "had not have",
    "hasn't": "has not",
    "haven't": "have not",
    "he'd": "he would",
    "he'd've": "he would have",
    "he'll": "he will",
    "he'll've": "he will have",
    "he's": "he is",
    "how'd": "how did",
    "how'd'y": "how do you",
    "how'll": "how will",
    "how's": "how is",
    "i'd": "I would",
    "i'd've": "I would have",
    "i'll": "I will",
    "i'll've": "I will have",
    "i'm": "I am",
    "i've": "I have",
    "isn't": "is not",
    "it'd": "it had",
    "it'd've": "it would have",
    "it'll": "it will",
    "it'll've": "it will have",
    "it's": "it is",
    "let's": "let us",
    "ma'am": "madam",
    "mayn't": "may not",
    "might've": "might have",
    "mightn't": "might not",
    "mightn't've": "might not have",
    "must've": "must have",
    "mustn't": "must not",
    "mustn't've": "must not have",
    "needn't": "need not",
    "needn't've": "need not have",
    "o'clock": "of the clock",
    "oughtn't": "ought not",
    "oughtn't've": "ought not have",
    "shan't": "shall not",
    "sha'n't": "shall not",
    "shan't've": "shall not have",
    "she'd": "she would",
    "she'd've": "she would have",
    "she'll": "she will",
    "she'll've": "she will have",
    "she's": "she is",
    "should've": "should have",
    "shouldn't": "should not",
    "shouldn't've": "should not have",
    "so've": "so have",
    "so's": "so is",
    "that'd": "that would",
    "that'd've": "that would have",
    "that's": "that is",
    "there'd": "there had",
    "there'd've": "there would have",
    "there's": "there is",
    "they'd": "they would",
    "they'd've": "they would have",
    "they'll": "they will",
    "they'll've": "they will have",
    "they're": "they are",
    "they've": "they have",
    "to've": "to have",
    "wasn't": "was not",
    "we'd": "we had",
    "we'd've": "we would have",
    "we'll": "we will",
    "we'll've": "we will have",
    "we're": "we are",
    "we've": "we have",
    "weren't": "were not",
    "what'll": "what will",
    "what'll've": "what will have",
    "what're": "what are",
    "what's": "what is",
    "what've": "what have",
    "when's": "when is",
    "when've": "when have",
    "where'd": "where did",
    "where's": "where is",
    "where've": "where have",
    "who'll": "who will",
    "who'll've": "who will have",
    "who's": "who is",
    "who've": "who have",
    "why's": "why is",
    "why've": "why have",
    "will've": "will have",
    "won't": "will not",
    "won't've": "will not have",
    "would've": "would have",
    "wouldn't": "would not",
    "wouldn't've": "would not have",
    "y'all": "you all",
    "y'alls": "you alls",
    "y'all'd": "you all would",
    "y'all'd've": "you all would have",
    "y'all're": "you all are",
    "y'all've": "you all have",
    "you'd": "you had",
    "you'd've": "you would have",
    "you'll": "you you will",
    "you'll've": "you you will have",
    "you're": "you are",
    "you've": "you have"
}

c_re = re.compile('(%s)' % '|'.join(c_dict.keys()))

add_stop = ['', ' ', 'say', 's', 'u', 'ap', 'afp', '...', 'n', '\\', 'think', 'test', 'need', 'make', 'look', 'good',
            'use', 'it, is', 'like', 'want', 'yes', 'try', 'work', 'just', 'html', 'org', 'sure', 'don', 'lot', 'didn',
            'doesn', 'http', 'com', 'know']

stop_words = ENGLISH_STOP_WORDS.union(add_stop)
tokenizer = TweetTokenizer()
pattern = r"(?u)\b\w\w+\b"
lemmatizer = WordNetLemmatizer()
punc = list(set(string.punctuation))


# Splits words on white spaces (leaves contractions intact) and splits out trailing punctuation
def casual_tokenizer(text):
    translator = str.maketrans(string.punctuation, ' ' * len(string.punctuation))
    text = text.translate(translator)
    tokens = tokenizer.tokenize(text)
    return tokens


# Function to replace the nltk pos tags with the corresponding wordnet pos tag to use the wordnet lemmatizer
def get_word_net_pos(treebank_tag):
    if treebank_tag.startswith('J'):
        return wordnet.ADJ
    elif treebank_tag.startswith('V'):
        return wordnet.VERB
    elif treebank_tag.startswith('N'):
        return wordnet.NOUN
    elif treebank_tag.startswith('R'):
        return wordnet.ADV
    else:
        return None


def lemma_wordnet(tagged_text):
    final = []
    for word, tag in tagged_text:
        wordnet_tag = get_word_net_pos(tag)
        if wordnet_tag is None:
            final.append(lemmatizer.lemmatize(word))
        else:
            final.append(lemmatizer.lemmatize(word, pos=wordnet_tag))
    return final


def expandContractions(text, c_re=c_re):
    def replace(match):
        return c_dict[match.group(0)]

    return c_re.sub(replace, text)


def convertMessagesToWordEmbeddings(embeddingsPath, df):
    word_embeddings = {}
    dimensions = 25
    messages = df['message']
    # Extract pre-trained word vectors from disk
    f = open(embeddingsPath, encoding='utf-8')
    for line in f:
        values = line.split()
        word = values[0]
        coefs = np.asarray(values[1:], dtype='float32')
        word_embeddings[word] = coefs
    f.close()

    messageEmbeddings = []
    for i in messages:
        if len(i) != 0:
            v = sum([word_embeddings.get(w, np.zeros((dimensions,))) for w in i.split()]) / (len(i.split()) + 0.001)
        else:
            v = np.zeros((dimensions,))
        messageEmbeddings.append(v)

    return messageEmbeddings


# function to remove stopwords
def remove_stopwords(message, stop_words):
    new = " ".join([i for i in message if i not in stop_words])
    return new


#Word Frequency With Bigrams
def word_freq_bigrams(clean_text_list, top_n):
    bigram_model = Phrases(clean_text_list, min_count=2, threshold=1)
    w_bigrams = list(bigram_model[clean_text_list])
    flat_w_bigrams = [item for sublist in w_bigrams for item in sublist]
    with_counts = Counter(flat_w_bigrams)
    top = with_counts.most_common(top_n)
    word = [each[0] for each in top]
    num = [each[1] for each in top]
    return pd.DataFrame([word, num]).T


def process_text(text):
    try:
        soup = BeautifulSoup(text)
        tags_del = soup.get_text()
        no_html = re.sub('<[^>]*>', '', tags_del)
        tokenized = casual_tokenizer(no_html)
        lower = [item.lower() for item in tokenized]
        decontract = [expandContractions(item, c_re=c_re) for item in lower]
        tagged = nltk.pos_tag(decontract)
        lemma = lemma_wordnet(tagged)
        no_num = [re.sub('[0-9]+', '', each) for each in lemma]
        no_punc = [w for w in no_num if w not in punc]
        no_stop = [w for w in no_punc if w not in stop_words]
        longer_than_2 = [w for w in no_stop if len(w) > 2]
        return longer_than_2
    except:
        print("text... %s", text)
        print("error")


def loadTextData(dataFile):
    df = pd.read_csv(dataFile)
    df['message'] = df['message'].apply(process_text)
    return df


# Evaluate baseline performance of various classifiers
def baseline_model_experiments(x, y):
    # Creating a dict of the models
    model_dict = {'Dummy': DummyClassifier(random_state=3),
                  'Stochastic Gradient Descent': SGDClassifier(random_state=3, loss='log'),
                  'Random Forest': RandomForestClassifier(random_state=3),
                  'Decision Tree': DecisionTreeClassifier(random_state=3),
                  'AdaBoost': AdaBoostClassifier(random_state=3),
                  'Multinomial Naive Bayes': MultinomialNB(),
                  'K Nearest Neighbor': KNeighborsClassifier(),
                  'SVM': SVC(random_state=3)
                  }

    # 10 fold cross validation experiments
    df = model_cv(model_dict, x, y)

    with pd.option_context('display.max_rows', None, 'display.max_columns', None):  # more options can be specified also
        print(df)


# Function to run 10-fold cross validation for each model in a dataframe
def model_cv(model_dict, x, y):
    scoring = {'acc': 'accuracy',
               'prec_macro': 'precision_macro',
               'rec_macro': 'recall_macro'}
    folds = 10
    cv_df = pd.DataFrame(index=range(folds * len(model_dict.keys())))
    entries = []
    for k, v in model_dict.items():
        if k != 'Multinomial Naive Bayes':
            print(k)
            results = cross_validate(v, x, y, scoring=scoring, cv=folds)  # Uses stratified cross validation by default
            # Predict the data using CV
            y_pred = cross_val_predict(v, x, y, cv=folds)
            # Get the confusion matrix and put it into a df
            cm = confusion_matrix(y, y_pred)
            print("cm for: %s", type(v).__name__)
            print(cm)
            for metric_name in results.keys():
                fold_idx = 0
                for item in results[metric_name]:
                    entries.append((k, metric_name, fold_idx, item))
                    fold_idx += 1
                average_score = np.average(results[metric_name])
                print('%s : %f' % (metric_name, average_score))

            cv_df = pd.DataFrame(entries, columns=['model_name', 'metric', 'fold_idx', 'value'])

    return cv_df


def postSplitSampling(model_dict, samplingName, samplingAlgorithm, X, y):
    entries = []
    results = []
    recalls = []
    precisions = []
    f1s = []
    for k, v in model_dict.items():
        print('Model: {k} with: {samplingName}')
        cv = KFold(n_splits=10, shuffle=False)
        fold_idx = 0
        for train_index, test_index in cv.split(X):
            X_train, X_test, y_train, y_test = X[train_index], X[test_index], y[train_index], y[test_index]

            # Apply sampling algorithm
            X_train_res, Y_train_res = samplingAlgorithm.fit_resample(X_train, y_train)
            v.fit(X_train_res, Y_train_res)
            item = v.score(X_test, y_test)
            results.append(item)

            # Obtain precision, recall and f1 scores
            # print(f'Fold {fold_idx} - X_test = {X_test.shape}')
            y_pred = v.predict(X_test)
            p = precision_score(y_test, y_pred, average='macro')
            r = recall_score(y_test, y_pred, average='macro')
            f1 = f1_score(y_test, y_pred, average='macro')
            precisions.append(p)
            recalls.append(r)
            f1s.append(f1)

            entries.append((k, 'accuracy', fold_idx, item))
            entries.append((k, 'precision', fold_idx, p))
            entries.append((k, 'recall', fold_idx, r))
            entries.append((k, 'f1', fold_idx, f1))
            fold_idx += 1

        print('%s : %f' % ('Accuracy', np.average(results)))
        print('%s : %f' % ('Precision', np.average(precisions)))
        print('%s : %f' % ('Recall', np.average(recalls)))
        print('%s : %f' % ('F1', np.average(f1s)))

    cv_df = pd.DataFrame(entries, columns=['model_name', 'metric', 'fold_idx', 'value'])
    return cv_df


# Sampling experiments
def sampling_experiments(x, y):
    sampling_strategies = {'SMOTE': SMOTE(),
                           'SVMSMOTE': SVMSMOTE(),
                           'BorderlineSMOTE': BorderlineSMOTE(),
                           'NCL': NeighbourhoodCleaningRule(),
                           'Tomek': TomekLinks(),
                           'SMOTETomek': SMOTETomek(),
                           }

    # # Creating a dict of the models
    model_dict = {'Dummy': DummyClassifier(random_state=3),
                  'Stochastic Gradient Descent': SGDClassifier(random_state=3, loss='log'),
                  'Random Forest': RandomForestClassifier(random_state=3),
                  'Decision Tree': DecisionTreeClassifier(random_state=3),
                  'AdaBoost': AdaBoostClassifier(random_state=3),
                  'Multinomial Naive Bayes': MultinomialNB(),
                  'K Nearest Neighbor': KNeighborsClassifier(),
                  'SVM': SVC(C=1.0, kernel='rbf', degree=3, coef0=0.0, random_state=3)
                  }

    for samplingName, sampling_algorithm in sampling_strategies.items():
        df = postSplitSampling(model_dict, samplingName, sampling_algorithm, x, y)
        with pd.option_context('display.max_rows', None, 'display.max_columns', None):
            print(df)


# Hyperparameter tunning
# Gridsearch with 10-fold cross validation
def hyperparameterSearch(X, y):

    random_states = [0, 3]

    # SGD
    loss = ['log', 'hinge']
    penalty = ['l2', 'l1', 'elasticnet']
    alpha = [1e-6, 1e-3, 1e-1, 1e0]
    max_iter = [5, 1000, 10000]
    epsilon = np.linspace(0.0, 1.0, 10, endpoint=True)
    tol = [None, 1e-3]
    lr = ['constant', 'optimal', 'invscaling', 'adaptive']
    eta0 = [0.1, 0.001, 0.5]
    
    clf = SGDClassifier()
    
    params = dict(loss=loss,
                  penalty=penalty,
                  alpha=alpha,
                  max_iter=max_iter,
                  tol=tol,
                  epsilon=epsilon,
                  learning_rate=lr,
                  eta0=eta0,
                  random_state=random_states)
    
    print("SGD GridSearch")
    gridsearch = GridSearchCV(clf, params, cv=10, verbose=1, n_jobs=-1)
    sgd_best_model_gs = gridsearch.fit(X, y)
    printGridSearchResults("SGD", sgd_best_model_gs)
    sgd_best_model = sgd_best_model_gs.best_estimator_  # SGDClassifier(**sgd_best_model_gs.best_params_)
    # sgd_best_model = SGDClassifier(alpha= 0.001, loss= 'log', max_iter= 5, penalty= 'l2', random_state= 3, tol= None)
    
    # Decision Tree
    max_depths = np.linspace(1, 100, 100, endpoint=True)  # fit a decision tree with depths ranging from 1 to 100
    max_features = ['auto', 'sqrt', 'log2']
    min_samples_splits = np.linspace(0.1, 1.0, 10, endpoint=True)
    min_samples_leaf = np.linspace(0.0, 1.0, 5, endpoint=True)
    criterions = ['gini', 'entropy']
    splitters = ['best', 'random']
    min_impurity_decreases = np.linspace(0.1, 1.0, 10, endpoint=True)
    clf = DecisionTreeClassifier()
    
    params = dict(max_depth=max_depths,
                  max_features=max_features,
                  min_samples_leaf=min_samples_leaf,
                  min_samples_split=min_samples_splits,
                  criterion=criterions,
                  splitter=splitters,
                  min_impurity_decrease=min_impurity_decreases,
                  random_state=random_states)
    
    print("DT GridSearch")
    gridsearch3 = GridSearchCV(clf,
                               params,
                               cv=10,
                               verbose=1,
                               n_jobs=-1)
    
    dt_best_model = gridsearch3.fit(X, y)
    printGridSearchResults("DT", dt_best_model)
    #
    # Random Forest
    max_depths = [10,20,30,40,50,60,70,80,90,100]  # fit a random forest with trees with depths ranging from 1 to 100
    max_features = ['auto', 'sqrt', 'log2']
    min_samples_splits = np.linspace(0.1, 1.0, 10, endpoint=True)
    min_samples_leaf = [2, 4, 6, 8, 10]
    n_estimators = [100,200,400,600, 800, 1000]
    criterions = ['gini', 'entropy']
    min_impurity_decreases = [0.1, 0.2, 0.3, 0.4,0.5]
    clf = RandomForestClassifier()

    params = dict(n_estimators=n_estimators,
                  criterion=criterions,
                  max_depth=max_depths,
                  max_features=max_features,
                  min_samples_leaf=min_samples_leaf,
                  min_samples_split=min_samples_splits,
                  min_impurity_decrease=min_impurity_decreases,
                  random_state=random_states)

    print("RF GridSearch")
    gridsearch5 = GridSearchCV(clf,
                               params,
                               cv=10,
                               verbose=1,
                               n_jobs=-1)

    rf_best_model = gridsearch5.fit(X, y)
    printGridSearchResults("RF", rf_best_model)
    
    # SVM
    Cs = np.linspace(0.001, 10, 10, endpoint=True)  # [0.001, 0.01, 0.1, 1, 10]
    degrees = [1,2,3,4,5,6,7,8,9,10]
    gammas = [0.001, 0.01, 0.1, 1]
    coefs = np.linspace(0.0, 1.0, 10, endpoint=True)
    kernels = ['rbf', 'sigmoid', 'linear']
    decision_function_shapes = ['ovo', 'ovr']
    shrinkings = [True, False]
    breaks = [True, False]
    tols = np.linspace(0.0, 1.0, 10, endpoint=True)
    caches = [100,200,300,400]
    
    clf = SVC()
    
    params = dict(C=Cs,
                  gamma=gammas,
                  degree=degrees,
                  coef0=coefs,
                  kernel=kernels,
                  shrinking=shrinkings,
                  tol=tols,
                  cache_size=caches,
                  decision_function_shape=decision_function_shapes,
                  break_ties=breaks,
                  random_state=random_states)
    
    print("SVM GridSearch")
    gridsearch2 = GridSearchCV(clf, params, cv=10, verbose=1, n_jobs=-1)
    
    svm_best_model_gs = gridsearch2.fit(X, y)
    printGridSearchResults("SVM", svm_best_model_gs)
    svm_best_model = svm_best_model_gs.best_estimator_

    AdaBoost
    rates = [0.01, 0.1, 0.5, 1]
    n_estimators = [200, 400, 600, 800, 1000, 1200, 1400, 1600, 1800, 2000]
    algorithms = ['SAMME', 'SAMME.R']
    
    clf = AdaBoostClassifier()
    
    params = dict(n_estimators=n_estimators,
                  learning_rate=rates,
                  algorithm=algorithms,
                  random_state=random_states)

    print("AdaBoost GridSearch")
    gridsearch3 = GridSearchCV(clf, params, cv=10, verbose=1, n_jobs=-1)
    
    ada_best_model_gs = gridsearch3.fit(X, y)
    printGridSearchResults("AdaBoost", ada_best_model_gs)
    ada_best_model = ada_best_model_gs.best_estimator_
    
    # NB
    alpha = np.linspace(0.0, 1.0, 50, endpoint=True)
    fit_priors = [True, False]
    
    clf = MultinomialNB()
    
    params = dict(alpha=alpha,
                  fit_prior=fit_priors)
    
    print("NB GridSearch")
    gridsearch4 = GridSearchCV(clf, params, cv=10, verbose=1, n_jobs=-1)
    
    nb_best_model_gs = gridsearch4.fit(X, y)
    printGridSearchResults("NB", nb_best_model_gs)
    nb_best_model = nb_best_model_gs.best_estimator_

    return rf_best_model # ada_best_model, nb_best_model#, svm_best_model  # , svm_best_model , dt_best_model


def printGridSearchResults(model, cv_gs_model):
    f = open("GridResults.txt", "w+")
    print("Best parameters set found:")
    f.write(model + ":\n")
    f.write("Best parameters set found:\n")
    print()

    print(cv_gs_model.best_params_)
    for key, value in cv_gs_model.best_params_.items():
        f.write('%s:%s\n' % (key, value))
    print()

    print("Best Accuracy achieved:")
    f.write("Best Accuracy achieved:\n")
    print()

    print(cv_gs_model.best_score_)
    f.write('%s\n' % cv_gs_model.best_score_)

    print("Grid scores tested:")
    f.write("Grid scores tested:\n")
    print()
    means = cv_gs_model.cv_results_['mean_test_score']
    stds = cv_gs_model.cv_results_['std_test_score']
    for mean, std, params in zip(means, stds, cv_gs_model.cv_results_['params']):
        print("%0.3f (+/-%0.03f) for %r" % (mean, std * 2, params))
        f.write("%0.3f (+/-%0.03f) for %r\n" % (mean, std * 2, params))
    print()
    f.close()


def main():
    datafile = 'AllchannelsForClass.csv'

    embeddingsPath = 'glove.twitter.27B/glove.twitter.27B.25d.txt'

    # loading the dataset
    print("Loading dataset")
    dataset = loadTextData(datafile)

    # Printing distribution information of the dataset
    LE = LabelEncoder()
    dataset['label_num'] = LE.fit_transform(dataset['category'])

    print("Total messages: %s", dataset['message'].count())
    print(dataset.groupby(['category'])['message'].count())
    print(dataset['category'].unique())
    print(dataset['label_num'].unique())

    # Preparing the dataframes

    # Splitting the df into the different categories
    df_devOps = dataset.loc[dataset['category'] == 'Dev-Ops']
    df_Comm = dataset.loc[dataset['category'] == 'Communication']
    df_Customer = dataset.loc[dataset['category'] == 'Customer support']
    df_Other = dataset.loc[dataset['category'] == 'Other']

    df_devOps = df_devOps.iloc[:]
    df_Comm = df_Comm.iloc[:]
    df_Customer = df_Customer.iloc[:]
    df_Other = df_Other.iloc[:]

    # Appending the dfs back together
    dataset = pd.concat([df_devOps, df_Comm, df_Customer, df_Other])
    df = dataset

    # Turning the labels into numbers
    df['label_num'] = LE.fit_transform(df['category'])

    display(df.groupby(['category'])['message'].count())
    display(df['category'].unique())
    display(df['label_num'].unique())

    text_devOps = df_devOps['message'].tolist()
    text_Comm = df_Comm['message'].tolist()
    text_Customer = df_Customer['message'].tolist()
    text_Other = df_Other['message'].tolist()

    # Creating the features (tf-idf weights) for the processed text

    messages = dataset['message'].astype('str')
    tfidf_vectorizer = TfidfVectorizer(ngram_range=(2, 3), stop_words='english', max_features=300)

    X = tfidf_vectorizer.fit_transform(messages)  # features
    y = dataset['label_num'].values  # target

    print(X.shape)
    print(y.shape)

    print('original dataset size: %s' % Counter(y))

    # Baseline models with default parameters
    print('Baseline models default parameters')
    baseline_model_experiments(X,y)

    # Hyperparameter tuning
    print('Hyperparameter tunning')
    hyperparameterSearch(X, y)

    # Defining sampling Models
    print('Sampling Experiments')
    sampling_experiments(X,y)


if __name__ == "__main__":
    main()