
import pandas as pd
import requests
import pycurl
import json
from io import BytesIO
import certifi


def getRoomIdByURI(token, chatRoomURI):
    myBuffer = BytesIO()
    headers = ['Authorization: Bearer {}'.format(token),
               'Content-Type: application/json',
               'Accept: application/json']

    curlConnect = pycurl.Curl()
    curlConnect.setopt(pycurl.CAINFO, certifi.where())
    curlConnect.setopt(pycurl.URL, 'https://api.gitter.im/v1/rooms')
    curlConnect.setopt(pycurl.POST, 1)
    curlConnect.setopt(pycurl.HTTPHEADER, headers)
    curlConnect.setopt(pycurl.POSTFIELDS, json.dumps({'uri': '{}'.format(chatRoomURI)}))
    curlConnect.setopt(pycurl.WRITEDATA, myBuffer)
    curlConnect.perform()

    print(curlConnect.getinfo(pycurl.HTTP_CODE))
    # --> 200
    print(curlConnect.getinfo(pycurl.EFFECTIVE_URL))
    # --> "https://www.python.org/"
    certinfo = curlConnect.getinfo(pycurl.INFO_CERTINFO)
    print("certinfo: {} ".format(certinfo))

    curlConnect.close()
    body = json.loads(myBuffer.getvalue().decode())
    room_id = body['id']
    print("{} room id:".format(chatRoomURI), room_id)
    return room_id

def readAllChannels(fileName,token):
    data = pd.read_csv(fileName)
    channels = []

    for index, row in data.iterrows():
        Community = row["Community"]
        room_name = row["Channel Name"]
        try:
            roomID = getRoomIdByURI(token, room_name)
            r=[Community,room_name,roomID]
            channels.append(r)
        except Exception as e:
            print("passing this", room_name)

    return channels


def getUsersInfo(room_id,token):
    users=[]
    while True:
        if len(users)==0:
            r = requests.get("https://api.gitter.im/v1/rooms/{}/users?access_token={}&limit=100".format(room_id, token))
            for i in r.json():
                print("username:", i["username"])
                users.append(i["username"])
        else:
            r = requests.get("https://api.gitter.im/v1/rooms/{}/users?access_token={}&limit=100&skip={}".format(room_id, token,len(users)))
            print("Skipping", len(users), "JSON:",len(r.json()))
            if(len(r.json())==0):
                break
            for i in r.json():
                print("username:", i["username"])
                users.append(i["username"])
    return users

def FidndevelopersInGithub(roomName):
    #https://api.github.com/repos/enzymejs/enzyme/contributors
    users = []
    pageNumber=1
    try:
        while True:
            #h="https://api.github.com/repos/{}/contributors?page={}&per_page={}$access_token={}".format(roomName,pageNumber,100,"write your token here")
            r = requests.get("https://api.github.com/repos/{}/contributors?page={}&per_page={}&access_token={}".format(roomName,pageNumber,100,"write your token here"))
            if len(r.json())==0 or not r.status_code == 200:
                break
            for i in r.json():
                print("github_user_name:", i["login"])
                users.append(i["login"])
            pageNumber+=1

    except Exception as e:
        print(e)
        print("No corresponding github was found ")
        return users

    return users

def intersection(lst1, lst2):
    return list(set(lst1) & set(lst2))

if __name__ == '__main__':
    token = ''
    #channels_info = readAllChannels(fileName="Gitter channels overview.csv",token=token) #based on the 139 channels only
    channels_info = readAllChannels(fileName="communities_details.csv",token=token)#based on all rooms in the 139 channels.
    roomIDs=[]
    column_name = ['Channel','roomName','# of Developers', '# of Github users','# of Gitter Users', 'Gitter Users List','Developer Users List','Developers in both']
    rows=[]
    for channel in channels_info:
        # r=[Community,room_name,roomID]
        community=channel[0]
        roomName=channel[1]
        roomID=channel[2]

        roomIDs.append(roomID)
        usersInRoom=getUsersInfo(roomID,token)
        developersInGithub=FidndevelopersInGithub(roomName=roomName)
        #TODO: get the intersection and then the length and save all data in a new CSV file.
        print(len(developersInGithub))
        developersGitterGithub=intersection(usersInRoom,developersInGithub)
        print(developersGitterGithub)
        print("Intersection: ",len(developersGitterGithub))
        rows.append((community,roomName,len(developersGitterGithub),len(developersInGithub),len(usersInRoom),usersInRoom, developersInGithub,developersGitterGithub))


    dev_df = pd.DataFrame(rows, columns=column_name)
    dev_df.to_csv("Developers per Channel2.csv", index=None)


