
from time import sleep
from gitterpy.client import GitterClient
import pandas as pd
import numpy as np
import os
import logging
from dateutil import parser
import sys
from collections import Counter

if not os.path.exists('gitter_token.txt'):
    print("Make a text file in the current dir and name it 'gitter_token.txt'. Then put your token inside the file.")
    exit()

with open("gitter_token.txt") as fd:
    lines = fd.readlines()

token_id = lines[0]

gitter = GitterClient(token_id)

def save_to_csv(df, count_files, file_name):
    # df.to_csv('output/output-{}.csv'.format(count_files), mode='a', header=True)
    name = str.replace(file_name, '/', '-')

    if not os.path.exists('output'):
        os.makedirs('output')

    if not os.path.exists('output/{}'.format(name)):
        os.makedirs('output/{}'.format(name))

    df.to_csv('output/{}/{}.csv'.format(name, count_files), header=True)

def save_room_info_to_csv(room):
    df = pd.DataFrame([room])

    if not os.path.exists('output-rooms.csv'):
        df.to_csv('output-rooms.csv', mode='a', header=True, index_label='#')
    else:
        df.to_csv('output-rooms.csv', mode='a', header=False)

def get_message_type(message_obj):
    if 'parentId' in message_obj:
        return 'reply'
    elif 'threadMessageCount' in message_obj and not np.isnan(message_obj['threadMessageCount']) :
        return 'topic'
    return 'message'

def does_contain_code(message):
    message = str(message)
    return 1 if message.find('<code>') > -1 else 0

def does_contain_quote(message):
    message = str(message)
    return 1 if message.find('"') > -1 or message.find("'") > -1 else 0



def form_message_data(df, room):

    if df.empty:
        return df

    new_df = pd.DataFrame()
    new_df['id'] = df['id']
    new_df['text'] = df['text']
    new_df['html'] = df['html'] if 'html' in df else df['text']
    new_df['sent'] = df['sent']

    new_df['sent_year'] = df['sent'].map(lambda sent: parser.parse(sent).year)
    new_df['sent_month'] = df['sent'].map(lambda sent: parser.parse(sent).month)
    new_df['sent_day'] = df['sent'].map(lambda sent: parser.parse(sent).day)
    new_df['sent_yearMonth'] = new_df.apply(lambda new_message: "{}".format(new_message['sent_year']) + "-" + "{0:2d}".format(new_message['sent_month']).replace(' ', '0'), axis=1)
    new_df['sent_yearMonthDay'] = new_df.apply(lambda new_message: "{}".format(new_message['sent_year']) + "-" + "{0:2d}".format(new_message['sent_month']).replace(' ', '0') + "-" + "{0:2d}".format(new_message['sent_day']).replace(' ', '0'), axis=1)

    new_df['editedAt'] = df['editedAt'] if 'editedAt' in df else ''
    new_df['readBy'] = df['readBy'] if 'readBy' in df else ''
    new_df['urls'] = df['urls'] if 'urls' in df else ''
    new_df['mentions'] = df['mentions'] if 'mentions' in df else ''
    new_df['meta'] = df['meta'] if 'meta' in df else ''
    new_df['v'] = df['v'] if 'v' in df else ''
    new_df['threadMessageCount'] = df['threadMessageCount'] if 'threadMessageCount' in df else ''
    new_df['parentId'] = df['parentId'] if 'parentId' in df else ''
    new_df['message_type'] = df.apply(get_message_type, axis=1)

    if 'fromUser' in df:
        new_df['fromUser_id'] = df['fromUser'].apply(lambda user: user['id'] if user is not np.nan else '')
        new_df['fromUser_username'] = df['fromUser'].apply(lambda user: user['username'] if user is not np.nan else '')
        new_df['fromUser_url'] = df['fromUser'].apply(lambda user: user['url'] if user is not np.nan and 'url' in user else '')
        new_df['fromUser_avatarUrl'] = df['fromUser'].apply(lambda user: user['avatarUrl'] if user is not np.nan and 'avatarUrl' in user else '')
    else:
        new_df['fromUser_id'] = ''
        new_df['fromUser_username'] = ''
        new_df['fromUser_url'] = ''
        new_df['fromUser_avatarUrl'] = ''



    new_df['room_group'] = room['group_name'] if 'group_name' in room else ''
    new_df['room_name'] = str.split(room['name'], '/')[1]
    new_df['room_full_name'] = room['name']

    new_df['contains_code'] = new_df['html'].map(does_contain_code)
    new_df['contains_quote'] = new_df['html'].map(does_contain_quote)

    return new_df

def get_topicType_messages(df):
    return df[df['message_type'] == 'topic']

def fetch_reply_messages(room, messages):
    topic_messages = get_topicType_messages(messages)
    all_reply_messages = pd.DataFrame()

    if len(topic_messages) > 0:
        print("Fetching replies for {} topics: ".format(len(topic_messages)))

        for index_topic, topic_message in topic_messages.iterrows():
            print('Fetching replies for {} ... '.format(topic_message['id']), end='')

            go_to_next = False
            temp_raw_data = None

            while not go_to_next:
                try:
                    temp_raw_data = gitter.messages.get('rooms/{}/chatMessages/{}/thread'
                                                        .format(room['id'], topic_message['id']))
                    go_to_next = True
                except Exception as e:
                    print('# EXCEPTION ', end='')
                    print(e)
                    print("#")
                    sleep(60)

            reply_messages = pd.DataFrame(temp_raw_data)
            reply_messages = form_message_data(reply_messages, room)
            all_reply_messages = all_reply_messages.append(reply_messages)
            print('Done')
            sleep(0.5)

    return all_reply_messages

def get_value_if_exists(df, key, default_value = ''):
    return df[key] if key in df else default_value

def form_room_data(room):
    room_dic = {}
    room_dic['id'] = room['id']
    room_dic['name'] = room['name']
    room_dic['topic'] = get_value_if_exists(room, 'topic')
    room_dic['avatarUrl'] = get_value_if_exists(room, 'avatarUrl')
    room_dic['uri'] = get_value_if_exists(room, 'uri')
    room_dic['userCount'] = get_value_if_exists(room, 'userCount')
    room_dic['unreadItems'] = get_value_if_exists(room, 'unreadItems')
    room_dic['mentions'] = get_value_if_exists(room, 'mentions')
    room_dic['lastAccessTime'] = get_value_if_exists(room, 'lastAccessTime')
    room_dic['lurk'] = get_value_if_exists(room, 'lurk')
    room_dic['url'] = get_value_if_exists(room, 'url')
    room_dic['githubType'] = get_value_if_exists(room, 'githubType')
    room_dic['security'] = get_value_if_exists(room, 'security')
    room_dic['tags'] = get_value_if_exists(room, 'tags')
    room_dic['roomMember'] = get_value_if_exists(room, 'roomMember')
    room_dic['groupId'] = get_value_if_exists(room, 'groupId')
    if 'groupId' in room:
        room_dic['group_id'] = get_value_if_exists(room['group'], 'id')
        room_dic['group_name'] = get_value_if_exists(room['group'], 'name')
        room_dic['group_uri'] = get_value_if_exists(room['group'], 'uri')
        room_dic['group_homeUri'] = get_value_if_exists(room['group'], 'homeUri')
        room_dic['group_backedBy_type'] = get_value_if_exists(room['group']['backedBy'], 'type')
        room_dic['group_backedBy_linkPath'] = get_value_if_exists(room['group']['backedBy'], 'linkPath')
        room_dic['group_avatarUrl'] = get_value_if_exists(room['group'], 'avatarUrl')
    room_dic['public'] = get_value_if_exists(room, 'public')
    room_dic['matrixRoomLink'] = get_value_if_exists(room, 'matrixRoomLink')
    room_dic['v'] = get_value_if_exists(room, 'v')

    return room_dic

def print_room_info(room):
    print("##------------------------ Room Info ------------------------##")
    print("Room ID:", room['id'])
    print("Room Name:", room['name'])
    print("Room Topic:", room['topic'])
    print("User Count:", room['userCount'])
    print("Url:", room['url'])
    print("Github Type:", room['githubType'])
    print("Tags:", room['tags'])
    print("Group Id:", room['groupId'])
    print("Group Name:", room['group_name'] if 'group_name' in room else '')
    print("###----------------------------------------------------------###")

def fetch_data(channel_name = 'laravel/laravel', channel_cat = None):
    count_records = 0
    count_files = 1
    last_date_in_the_room = None
    first_date_in_the_room = None

    room = gitter.rooms.grab_room(channel_name)

    if 'error' in room:
        print("[{}] Failed to read room info: {}".format(channel_name, room['error']))
        return

    room = form_room_data(room)
    room_id = room['id']

    # Print room info
    print_room_info(room)

    # Get latest 50 messages
    print('{} ... '.format(count_files), end='')

    messages = gitter.messages.get('rooms/{}/chatMessages?limit=100'.format(room_id))

    messages = pd.DataFrame(messages)
    messages = form_message_data(messages, room)

    if len(messages) > 0:
        first_id_in_messages = messages.iloc[0]['id']

        first_date_in_the_room = messages.iloc[0]['sent_yearMonthDay']
        last_date_in_the_room = messages.iloc[-1]['sent_yearMonthDay']


        # Check topics and fetch replies
        replies = fetch_reply_messages(room, messages)
        messages = messages.append(replies)

        # Save
        save_to_csv(messages, count_files, channel_name)
        count_files += 1

        print('Done')

        count_records += len(messages)

        print("@@", channel_name, "@@", first_id_in_messages,
              "@@ {}/{}".format(messages.iloc[0]['sent_year'], messages.iloc[0]['sent_month']), "@@ Total:", count_records, end=" ")


        # Check if there are more messages to read
        while len(messages) >= 100:
            print(count_files, " ... ", end='')
            go_to_next = False

            while not go_to_next:
                try:
                    temp_raw_data = gitter.messages.get('rooms/{}/chatMessages?limit=100&beforeId={}'
                                                        .format(room_id, first_id_in_messages))

                    messages = pd.DataFrame(temp_raw_data)

                    # if there is no messages, break
                    if len(messages) <= 0:
                        break

                    messages = form_message_data(messages, room)

                    first_id_in_messages = messages.iloc[0]['id']
                    first_date_in_the_room = messages.iloc[0]['sent_yearMonthDay']

                    # Check topics and fetch replies
                    replies = fetch_reply_messages(room, messages)
                    if len(replies) > 0:
                        messages = messages.append(replies)

                    count_records += len(messages)

                    print("@@", channel_name,"@@" , first_id_in_messages, "@@ {}/{}".format(messages.iloc[0]['sent_year'], messages.iloc[0]['sent_month']) , "@@ Total:", count_records)
                    go_to_next = True
                except Exception as e:
                    print('.', end='')
                    logging.warning(e)
                    sleep(60)
            save_to_csv(messages, count_files, channel_name)
            count_files += 1
            print(".. Done")
            sleep(0.5)


    print("##########################")
    print(channel_name)
    print("Done Fetching All Messages")
    print("Total: {}".format(count_records))
    print("##########################")

    room['total_messages'] = count_records
    room['first_message_date'] = first_date_in_the_room
    room['last_message_date'] = last_date_in_the_room
    room['room_category'] = channel_cat
    save_room_info_to_csv(room)
    print("Room info has been saved.")

if __name__ == '__main__':
    #sys.argv=["","laravel/laravel"]
    if len(sys.argv) == 2:
        channel_name = sys.argv[1].strip()
        print("Reading the room: {}".format(channel_name))
        fetch_data(channel_name)
    else:
        file_name = "rooms.txt"

        print("Reading from the file ({}) ...".format(file_name))

        if not os.path.exists(file_name):
            print("Make a text file in the current dir and name it '{}'. Then put each room name in a new line inside the file.".format(file_name))
            exit()

        with open(file_name) as fd:
            lines = fd.readlines()
            count = 0
            channel_cat = ''
            for line in lines:
                count += 1
                channel_name = line.strip()
                if "/" in line:
                    print(count, "=>", channel_name)
                    fetch_data(channel_name, channel_cat)
                else:
                    channel_cat = line.strip()