library(stats)
library(effsize)


setwd("/Documents/Research/")

data <- read.csv("Results.csv", stringsAsFactors = FALSE)

#print(data)
# To determine the interpretation of a value for effect size
cliffs_inter <- function(x){
  x <- abs(x)
  if(x < 0.14){
    return("negligible")
  }
  
  if(x >= 0.14 && x < 0.33){
    return("small")
  }
  
  if(x >= 0.33 && x < 0.474){
    return("medium")
  }
  
  if(x >= 0.474) {
    return("large")
  }
  
  return ("huge*")
}

classifiers <- unique(data$approach)
samplings <- unique(data$targetArea)
print(classifiers)
print(samplings)


results <- matrix(0,nrow = 0, ncol = 7);
		
		nonh <- subset(data, data$approach == "NoHeuristics" & data$targetArea == "editingWindow", select = accuracy)
		nonh <- as.numeric(unlist(nonh))
        other <- subset(data, data$approach == "Heuristics" & data$targetArea == "editingWindow", select = accuracy)
        other <- as.numeric(unlist(other))
        
        if(length(other) != 550){
          message("something wrong")
        }
        
        print(nonh)
        print(other)
        
        test <- wilcox.test(nonh,other)
        cliff <- cliff.delta(nonh,other)
        estimate <- cliff$estimate
        inter <- cliffs_inter(cliff$estimate)
        
        results <- rbind(results, c("RandomForest-NCL-correlation",paste(classifier,sampling,sep = "-"),test$p.value,test$p.value<=0.05,test$p.value<=0.01,estimate,inter))

colnames(results) <- c("Best","Other","pvalue","isH195","isH199","cliff","effect")
results <- as.data.frame(results, stringsAsFactors = FALSE)
results$pvaluef <- p.adjust(results$pvalue, method = "bonferroni")
results$isH1f <- results$pvaluef<=0.01

write.csv(results,"stat_test_classifiers(editing).csv",row.names = FALSE,quote = FALSE)
